// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::constants {
    // precision of the coordinates
    // float is used for better iteration performance
    using coords_precision_t = double;
}