// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

#include <math/MathConcepts.h>

namespace ausaxs {
    template<numeric T>
    class Matrix;

    template<numeric T>
    class Vector;

    template<numeric T>
    class Vector3;
}