// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

#include <mini/detail/Evaluation.h>
#include <mini/detail/Parameter.h>
#include <mini/detail/FittedParameter.h>
#include <mini/detail/Landscape.h>
#include <mini/detail/Result.h>
#include <mini/MinimizerFactory.h>