// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::residue {
    class ResidueStorage;
    namespace detail {
        struct Atom;
        struct Bond;
        class Residue;
    }
}