// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::em::detail::header {
    class IMapHeader;
    class MRCHeader;
    class RECHeader;
    class DummyHeader;
}