// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::settings {
    /**
     * @brief Validate and correct the current settings. 
     */
    void validate_settings();
}