// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::io {
    class File;
    class ExistingFile;
    class Folder;

    namespace detail {
        struct Reader;
        struct Writer;
    }
}