// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::em {
	class ImageStackBase;

	namespace detail {
		struct ExtendedLandscape;
	}
}

namespace ausaxs::em::managers {
	class ProteinManager;
}