// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

#include <rigidbody/sequencer/ConvertSymmetryElement.h>

#include <cassert>

void ausaxs::rigidbody::sequencer::ConvertSymmetryElement::run() {
    assert(false && "ConvertSymmetryElement::run: Not implemented.");
}