// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::container {
    template<typename T> class Container1D;
    template<typename T> class Container2D;
    template<typename T> class Container3D;
}