// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#include <em/detail/header/data/RECData.h>

using namespace ausaxs;

em::detail::header::RECData::RECData() = default;
em::detail::header::RECData::~RECData() = default;
