// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

#include <settings/CrystalSettings.h>
#include <settings/EMSettings.h>
#include <settings/ExvSettings.h>
#include <settings/FitSettings.h>
#include <settings/Flags.h>
#include <settings/GeneralSettings.h>
#include <settings/GridSettings.h>
#include <settings/HistogramSettings.h>
#include <settings/PlotSettings.h>
#include <settings/MoleculeSettings.h>
#include <settings/RigidBodySettings.h>
#include <settings/SettingsIO.h>
#include <settings/SettingsValidation.h>