// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

// This also forwards some elements of data/record for simplicity
namespace ausaxs::data {
    class Body;
    class Molecule;
    struct Atom;
    class AtomFF;
    struct Water;
}