// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#include <em/detail/header/data/MRCData.h>

using namespace ausaxs;

em::detail::header::MRCData::MRCData() = default;
em::detail::header::MRCData::~MRCData() = default;
