// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::rigidbody::sequencer {
    class Sequencer;
    class ConstraintIteratorElement;
    class ParameterElement;
    class BodySelectElement;
    class TransformElement;
    class LoopElement;
    class EveryNStepElement;
    class SaveElement;
    class OptimizeStepElement;
}