// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::constants {
    inline constexpr char version[] = "v1.1.8";
    inline constexpr char licence[] = 
        "Copyright (C) 2025 Kristian Lytje & Jan Skov Pedersen.\n"
        "This program comes with ABSOLUTELY NO WARRANTY.\n"
        "This is free software, and you are welcome to redistribute it under certain conditions.\n"
        "For details see the LICENCE file.\n";
}