// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::fitter {
    class Fitter;
    class LinearFitter;
    class SmartFitter;
    class FitResult;
}