// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::hydrate {
    class Hydration;
    class HydrationStrategy;
    class ImplicitHydration;
    class ExplicitHydration;
    class NoHydration;
}