// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#include <hydrate/generation/NoHydration.h>
#include <hydrate/ExplicitHydration.h>
#include <data/atoms/Water.h>

using namespace ausaxs;

void hydrate::NoHydration::hydrate() {}