// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

// convenience header

#pragma once

#include <plots/Plot.h>
#include <plots/PlotDataset.h>
#include <plots/PlotHistogram.h>
#include <plots/PlotDistance.h>
#include <plots/PlotProfiles.h>
#include <plots/PlotLandscape.h>
#include <plots/PlotOptions.h>