// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::io::pdb {
    enum class RecordType;
    class PDBStructure;
    class PDBAtom;
    class PDBWater;
    class Footer;
    class Header;
    class Record;
    class Terminate;    
}