// SPDX-License-Identifier: LGPL-3.0-or-later
// Author: Kristian Lytje

#pragma once

namespace ausaxs::hist {
    class Distribution1D;
    class Distribution2D;
    class Distribution3D;
    class WeightedDistribution1D;
    class WeightedDistribution2D;
    class WeightedDistribution3D;
}